/*
 *	x11.c - Xlib interface for Euphoric
 *	AYM 2000-12-30
 */

/*
This file is copyright Andr Majorel 2000-2003.

This program is free software; you can redistribute it and/or modify it under
the terms of version 2 of the GNU General Public License as published by the
Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place, Suite 330, Boston, MA 02111-1307, USA.
*/


#include "configure.h"

#include <assert.h>			/* Standard C */
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <unistd.h>			/* Unix-specific */
#if HAVE_XSHM
#  include <sys/ipc.h>
#  include <sys/shm.h>
#endif

#include <X11/Xlib.h>			/* X11-specific */
#include <X11/Xutil.h>
#if HAVE_XSHM
#  include <X11/extensions/XShm.h>
#endif

#include "xeuphoric.h"
#include "hosttraps.h"
#include "stats.h"


typedef unsigned long pixel_t;

/* FIXME use headers */
extern volatile unsigned char Tick;

static int x11_error_handler (Display *dpy, XErrorEvent *e);
static void x11_exit (void);

Display *dpy;
int scn;
Window win;
int depth;
int width;
int height;
Visual *vis;
int visclass;
int viscolours;
Colormap cmap = 0;
GC gc;
XImage *img;
int linepad;
int bigendian;
int xputimage_in_progress;
int expose;
static pixel_t colours[COLOURS + 1];
static int tracebuf[VIDEO_WIDTH];
static size_t tracehead = 0;
int zoom = 1;
xshmpolicy_t xshmpolicy = XSHMPOLICY_AUTO;
#if HAVE_XSHM
static int shmid = -1;			/* ID of our shared memory segment */
static void *shmaddr = NULL;		/* Where it's attached */
static int xshmattach = 0;		/* XShmAttach() succeeded */
static int xshm;			/* img is a shared XImage */
static XShmSegmentInfo xshminfo;
#endif
static int x11_error;
static XErrorEvent x11_error_event;


void x11_start (void)
{
  /* Install the automatic clean-up code */
#if HAVE_XSHM
  shmid = -1;
  shmaddr = NULL;
  xshmattach = 0;
#endif
  hosttraps_register (x11_exit);

  /* Open connection with X server */
  dpy = XOpenDisplay (NULL);
  if (dpy == NULL) {
    err ("can't open display");
    exit (1);
  }

  /* Which screen ? */
  scn = DefaultScreen (dpy);
  linepad = BitmapPad (dpy);
  bigendian = ImageByteOrder (dpy) == MSBFirst;
  if (bigendian) {
    err ("big-endian X server--not supported yet");
    exit (1);
  }

  /* Create window */
  width = VIDEO_WIDTH * zoom;
  height = VIDEO_HEIGHT * zoom;
  win = XCreateSimpleWindow (dpy, DefaultRootWindow (dpy), 10, 10,
      width, height, 0, 0, 0);
  {
    XWindowAttributes wa;
    XGetWindowAttributes (dpy, win, &wa);
    depth = wa.depth;
    vis = wa.visual;
    /* Retrieve the attributes of the window's visual */
    {
      XVisualInfo vismodel;
      XVisualInfo *vislist;
      int nvisuals;

      vismodel.visualid = XVisualIDFromVisual (vis);
      vislist = XGetVisualInfo (dpy, VisualIDMask, &vismodel, &nvisuals);
      if (vislist == NULL)
      {
	err ("XGetVisualInfo() returned NULL");
	exit (1);
      }
      if (nvisuals < 1)
      {
	err ("XGetVisualInfo() returned less than one visual");
	exit (1);
      }
      visclass = vislist->class;
      viscolours = vislist->colormap_size;
    }
  }
  XSelectInput (dpy, win,
    KeyPressMask|KeyReleaseMask
    |ButtonPressMask|ButtonReleaseMask
    |PointerMotionMask
    |EnterWindowMask|LeaveWindowMask
    |ExposureMask
    |StructureNotifyMask);

  /* Set window title */
  {
    static const char self[] = "Xeuphoric ";
    char *buf = malloc (strlen (self) + strlen (version) + 1);
    if (buf == NULL)
    {
      err ("not enough memory");
      exit (1);
    }
    strcpy (buf, self);
    strcat (buf, version);
    XStoreName (dpy, win, buf);
    free (buf);
  }

  /* Set window size hints */
  {
    XSizeHints *hints = XAllocSizeHints ();
    hints->flags = PMinSize|PMaxSize|PBaseSize;
    hints->min_width = width;
    hints->min_height = height;
    hints->max_width = width;
    hints->max_height = height;
    hints->base_width = width;
    hints->base_height = height;
    XSetWMNormalHints (dpy, win, hints);
  }

  /* Create GC */
  {
    XGCValues gcv;
    unsigned long mask;
    mask = GCForeground|GCFunction|GCLineWidth;
    gcv.foreground = WhitePixel (dpy, scn);
    gcv.background = BlackPixel (dpy, scn);
    gcv.function = GXcopy;
    gc = XCreateGC (dpy, win, mask, &gcv);
    if (gc == NULL)
    {
      err ("can't create GC");
      exit (1);
    }
  }

  /* Create the XImage. The order of the steps is carefully chosen to
     facilitate the implementation of the -X auto policy (i.e. falling
     back on regular XImages when shared XImages are not available). */

  /* 1. Does the X server support the XShm extension ? */
  if (xshmpolicy == XSHMPOLICY_AUTO || xshmpolicy == XSHMPOLICY_ALWAYS)
  {
#if !HAVE_XSHM
    if (xshmpolicy == XSHMPOLICY_ALWAYS)
    {
      err ("this Xeuphoric binary was compiled without XShm support");
      exit (1);
    }
#else
    xshm = 0;
    if (XShmQueryExtension (dpy) == False)
    {
      err ("XShm extension not present on X server");
      if (xshmpolicy == XSHMPOLICY_ALWAYS)
	exit (1);
    }
    else
      xshm = 1;
#endif
  }

  /* 2. Create the XImage (shared or not). In my experience
	XShmCreateImage never fails, even if the server is remote, so
	don't bother with fallback here. */
  img = NULL;
#if HAVE_XSHM
  if (xshm)
  {
    img = XShmCreateImage (dpy, vis, depth, ZPixmap, 0, &xshminfo,
	width, height);
    if (img == NULL)
    {
      err ("XShmCreateImage failure");
      if (xshmpolicy == XSHMPOLICY_ALWAYS)
	exit (1);
      xshm = 0;
    }
  }
#endif
  if (img == NULL)
  {
    img = XCreateImage (dpy, vis, depth, ZPixmap, 0, 0,
		    width, height, linepad, 0);
    if (img == NULL)
    {
      err ("XCreateImage failure");
      exit (1);
    }
  }

  /* 3. Allocate the XImage buffer. If XShm appears to be supported,
	create a shared memory segment and call XShmAttach. If the X
	server is remote this will fail with BadAlloc and this is where
	we fall back on plain XImages. */
#if HAVE_XSHM
  if (xshm)
  {
    shmid = shmget (IPC_PRIVATE, img->bytes_per_line * height, IPC_CREAT|0777);
    if (shmid == -1)
    {
      err ("shmget: %s", strerror (errno));
      exit (1);
    }

    shmaddr = shmat (shmid, NULL, 0);
    if (shmaddr == NULL)
    {
      err ("shmat: %s", strerror (errno));
      exit (1);
    }

    xshminfo.shmid    = shmid;
    xshminfo.shmaddr  = shmaddr;
    xshminfo.readOnly = False;
    x11_error = 0;
    XSetErrorHandler (x11_error_handler);
    XSynchronize (dpy, True);
    xshmattach = 0;
    if (XShmAttach (dpy, &xshminfo) != True || x11_error)
    {
      char buf[100];

      XGetErrorText (dpy, x11_error_event.error_code, buf, sizeof buf);
      err ("XShmattach: %s", buf);

      /* Release the shared memory segment to avoid ugly messages from
         x11_exit(). */
      if (shmdt (shmaddr) != 0)
	err ("shmdt: %s", strerror (errno));
      shmaddr = NULL;			/* FIXME race condition */
      if (shmctl (shmid, IPC_RMID, NULL) != 0)
	err ("shmctl: %s", strerror (errno));
      shmid = -1;			/* FIXME race condition */

      if (xshmpolicy == XSHMPOLICY_ALWAYS)
	exit (1);

      xshm = 0;				/* Fall back on plain XImages */
      XDestroyImage (img);
      img = XCreateImage (dpy, vis, depth, ZPixmap, 0, 0,
		      width, height, linepad, 0);
      if (img == NULL)
      {
	err ("XCreateImage failure");
	exit (1);
      }
    }
    else
      xshmattach = 1;			/* FIXME race condition */
    XSynchronize (dpy, False);
    XSetErrorHandler (NULL);
  }
#endif

  /* 4. Set img->data */
#if HAVE_XSHM
  if (xshm)
  {
    img->data = shmaddr;
  }
  else
#endif
  {
    img->data = malloc (img->bytes_per_line * height);
    if (img->data == NULL)
    {
      err ("can't allocate XImage buffer (%s)", strerror (ENOMEM));
      exit (1);
    }
  }

  /* 5. We now have an XImage, one way or another */
  if (img->bits_per_pixel != 1
      && img->bits_per_pixel != 8
      && img->bits_per_pixel != 16
      && img->bits_per_pixel != 24
      && img->bits_per_pixel != 32)
  {
    err ("XImage: unsupported bits_per_pixel %d (not 1, 8, 16, 24 or 32)",
	img->bits_per_pixel);
    exit (1);
  }


  /* True colour type visual - no allocation is needed */
  if (visclass == TrueColor)
  {
    /* Pixel values for the 8 colours of the Oric plus grey */
    {
      int c;

      for (c = 0; c < sizeof colours / sizeof *colours; c++)
      {
	if (c == COLOURS)		/* Colour 8 = grey */
	  colours[c] = (img->red_mask   & ~(img->red_mask   >> 1))
		     | (img->green_mask & ~(img->green_mask >> 1))
		     | (img->blue_mask  & ~(img->blue_mask  >> 1));
	else
	{
	  colours[c] = 0;
	  if (c & 1)
	    colours[c] |= img->red_mask;
	  if (c & 2)
	    colours[c] |= img->green_mask;
	  if (c & 4)
	    colours[c] |= img->blue_mask;
	}
      }
    }

    /* On 24 bpp visuals, there are just 24 useful bits out of
       32. Make sure they are aligned with the start address of
       the pixel_t, so that memcpy()'ing a pixel_t onto an
       XImage does the right thing. */
    if (img->bits_per_pixel == 24)
    {
      pixel_t v = (img->red_mask | img->green_mask | img->blue_mask);
      unsigned char *p = (unsigned char *) &v;

      if (p[0] == 0xff && p[1] == 0xff && p[2] == 0xff && p[3] == 0x00)
	;				/* Already aligned */
      else if (p[0] == 0x00 && p[1] == 0xff && p[2] == 0xff && p[3] == 0xff)
      {
	int c;

	for (c = 0; c < sizeof colours / sizeof *colours; c++)
	{
	  unsigned char *p = (unsigned char *) (colours + c);

	  p[0] = p[1];
	  p[1] = p[2];
	  p[2] = p[3];
	  p[3] = 0;
	}
      }
      else
	err ("warning: strange RGB mask %08lXh, expect incorrect colours");
    }
  }

  /* Palette type visual - first try to satisfy our needs with
     read-only cells. If that fails, it means that the palette
     has no free cells left so fall back on nearest existing
     palette entries. */
  else if (visclass == PseudoColor || visclass == StaticColor
      || visclass == GrayScale || visclass == StaticGray
      || visclass == DirectColor)
  {
    int c;
    int satisfied[sizeof colours / sizeof *colours];
    int unsatisfied = 0;

    cmap = DefaultColormap (dpy, scn);
    for (c = 0; c < sizeof colours / sizeof *colours; c++)
    {
      XColor wanted;

      if (c == COLOURS)			/* Colour 8 = grey */
      {
	wanted.red   = 0x8000;
	wanted.green = 0x8000;
	wanted.blue  = 0x8000;
	wanted.flags = DoRed | DoGreen | DoBlue;
      }
      else
      {
	wanted.red   = (c & 1) ? 0xffff : 0;
	wanted.green = (c & 2) ? 0xffff : 0;
	wanted.blue  = (c & 4) ? 0xffff : 0;
	wanted.flags = DoRed | DoGreen | DoBlue;
      }
      if (XAllocColor (dpy, cmap, &wanted) == True)
      {
	colours[c] = wanted.pixel;
	satisfied[c] = 1;
      }
      else
      {
	satisfied[c] = 0;
	unsatisfied++;
      }
    }
    if (unsatisfied != 0)
    {
      XColor *palette;
      int palsize;

      /* Retrieve the palette */
      palsize = viscolours;
      palette = malloc (palsize * sizeof *palette);
      if (palette == NULL)
      {
	err ("not enough memory");
	exit (1);
      }
      for (c = 0; c < palsize; c++)
	palette[c].pixel = c;
      XQueryColors (dpy, cmap, palette, palsize);

      /* For each unsatisfied colour, use the closest existing
         colour in the palette. */
      for (c = 0; c < sizeof colours / sizeof *colours; c++)
      {
	int i;
	int besti = 0;
	double bestdistance = 3;
	double r = (c & 1) ? 1.0 : 0.0;
	double g = (c & 2) ? 1.0 : 0.0;
	double b = (c & 4) ? 1.0 : 0.0;

	if (satisfied[c])
	  continue;

	if (c == COLOURS)		/* Colour 8 = grey */
	{
	  r = 0.5;
	  g = 0.5;
	  b = 0.5;
	}
	else
	{
	  r = (c & 1) ? 1.0 : 0.0;
	  g = (c & 2) ? 1.0 : 0.0;
	  b = (c & 4) ? 1.0 : 0.0;
	}

	for (i = 0; i < palsize; i++)
	{
	  double dr = (double) palette[i].red   / 0xffff - r;
	  double dg = (double) palette[i].green / 0xffff - g;
	  double db = (double) palette[i].blue  / 0xffff - b;
	  double distance = dr * dr + dg * dg + db * db;
	  if (distance < bestdistance)
	  {
	    besti = i;
	    bestdistance = distance;
	  }
	}
	colours[c] = palette[besti].pixel;
	err ("warning: colour #%d: using rgb:%02X/%02X/%02X",
	    c,
	    palette[besti].red   >> 8,
	    palette[besti].green >> 8,
	    palette[besti].blue  >> 8);
      }
    }
  }
  else
  {
    err ("unknown visual class %d");
    exit (1);
  }

  /*XSynchronize (dpy, True);*/
  /* Wait for expose event */
  XMapWindow (dpy, win);
  XSync (dpy, False);
  for (;;)
  {
    XEvent ev;
    XNextEvent (dpy, &ev);
    if (ev.type == Expose)
      break;
  }
}

void x11_end (void)
{
#if HAVE_XSHM
  if (xshmattach)
  {
    if (XShmDetach (dpy, &xshminfo) != True)
      err ("XShmDetach failure");
    xshmattach = 0;			/* FIXME race condition */
  }
#endif

  XDestroyImage (img);

#if HAVE_XSHM
  XSync (dpy, False);			/* Detach the X server before RMID */

  if (shmaddr != NULL)
  {
    if (shmdt (shmaddr) != 0)
      err ("shmdt: %s", strerror (errno));
    shmaddr = NULL;			/* FIXME race condition */
  }

  if (shmid != -1)
  {
    if (shmctl (shmid, IPC_RMID, NULL) != 0)
      err ("shmctl: %s", strerror (errno));
    shmid = -1;				/* FIXME race condition */
  }
#endif

  XFreeGC (dpy, gc);
  XDestroyWindow (dpy, win);
  XCloseDisplay (dpy);
}

int x11_key (XEvent *ev)
{
  do
  {
    if (XPending (dpy) == 0)
      return 0;
    XNextEvent (dpy, ev);
    if (ev->type == Expose)
      expose = 1;
  }
  while (ev->type != KeyPress && ev->type != KeyRelease);
  return 1;
}


void display_frame (void)
{
  int y0, y1;				/* First and last scanlines touched */

#if 0
  if (Tick != 0)
	  return;
#endif
  stats.frames++;
  x11_trace (Tick);

  render_frame ();

  y0 = -1;
  y1 = 224;			/* Silence the "uninitialised" warning */
  if (frametouched)
  {
    if (img->bits_per_pixel == 1)
    {
      if (img->bitmap_bit_order == LSBFirst)
      {
#undef STRAIGHT_BITMAP
#define STRAIGHT_BITMAP 0
#include "x11render1bpp.c"
      }
      else if (img->bitmap_bit_order == MSBFirst)
      {
#undef STRAIGHT_BITMAP
#define STRAIGHT_BITMAP 1
#include "x11render1bpp.c"
      }
      else
      {
	err ("invalid bitmap_bit_order %d", (int) img->bitmap_bit_order);
      }
    }
    else if (img->bits_per_pixel == 8)
    {
      typedef unsigned char ximage_pixel_t;
#include "x11render.c"  /* If this was C++, we would use a template */
      assert (sizeof (ximage_pixel_t) == 1);
    }
    else if (img->bits_per_pixel == 16)
    {
      typedef unsigned short ximage_pixel_t;
#include "x11render.c"  /* If this was C++, we would use a template */
      assert (sizeof (ximage_pixel_t) == 2);
    }
    else if (img->bits_per_pixel == 24)
    {
      typedef unsigned char ximage_pixel_t[3];
#define USE_MEMCPY
#include "x11render.c"  /* If this was C++, we would use a template */
#undef USE_MEMCPY
      assert (sizeof (ximage_pixel_t) == 3);
    }
    else if (img->bits_per_pixel == 32)
    {
      typedef unsigned long ximage_pixel_t;
#include "x11render.c"  /* If this was C++, we would use a template */
      assert (sizeof (ximage_pixel_t) == 4);
    }
    else				/* Can't happen */
    {
      err ("unsupported bits_per_pixel %d", (int) img->bits_per_pixel);
    }
  }

  if (frametouched || expose)
  {
    int y;
    int h;

    if (expose)			/* Expose event: always redraw everything */
    {
      y = 0;
      h = height;
    }
    else
    {
      y = zoom * y0;
      h = zoom * (y1 + 1 - y0);
    }
    xputimage_in_progress = 1;	/* Don't be interrupted by SIGALRM */
#if HAVE_XSHM
    if (xshm)
      XShmPutImage (dpy, win, gc, img, 0, y, 0, y, width, h, False);
    else
#endif
      XPutImage (dpy, win, gc, img, 0, y, 0, y, width, h);
    XFlush (dpy);
    expose = 0;
    xputimage_in_progress = 0;
    stats.scanlines += h;
  }

#if 0
  {
    static int frames = 0;
    static time_t timeprev = 0;
    time_t timenow;
    frames++;
    time (&timenow);
    if (timenow != timeprev)
    {
      if (timeprev != 0)
	err (" fps = %d", (int) (frames / (timenow - timeprev)));
      frames = 0;
      timeprev = timenow;
    }
  }
#endif

  /* FIXME this is a quick and dirty workaround. The effect is to
     wait for the next SIGARLM. If we didn't, Xeuphoric would run
     at full tilt. I feel that the clean solution involves
     modifying the 6502 emulation code to return after
     1E6/refresh_rate cycles instead of calling back
     display_frame(). Said code has so far successfully countered
     my attempts at wrapping my head around it.

     This trick is doubly dirty because it relies on an
     implementation detail, which is that usleep() be implemented
     using SIGALRM. This is the case on Linux with Glibc 2.2. On
     implementations where SIGALARM doesn't interrupt usleep(),
     bad things will happen, like Xeuphoric running too slow or
     worse. */
  usleep (20000);    /* max(1/50 Hz, 1/60 Hz) = 20,000 s */
}

void Reset_Screen (void)
{
  while (Tick != 1)
    ;
  Tick = 0;
  display_frame ();
}

void x11_trace (int value)
{
  tracebuf[tracehead++] = value;
  if (tracehead >= sizeof tracebuf / sizeof *tracebuf)
    tracehead = 0;
}

static int x11_error_handler (Display *dpy, XErrorEvent *e)
{
  x11_error = 1;
  x11_error_event = *e;
  return 42;  /* Not used */
}

/*
 *	x11_exit - atexit() callback
 *
 *	Make sure we release our shared memory segment before we die.
 */
static void x11_exit (void)
{
#if HAVE_XSHM
  if (xshmattach)
  {
    if (XShmDetach (dpy, &xshminfo) != True)
      err ("XShmDetach failure");
    xshmattach = 0;
  }

  if (shmaddr != NULL)
  {
    if (shmdt (shmaddr) != 0)
      err ("shmdt: %s", strerror (errno));
    shmaddr = NULL;
  }

  if (shmid != -1)
  {
    if (shmctl (shmid, IPC_RMID, NULL) != 0)
      err ("shmctl: %s", strerror (errno));
    shmid = -1;
  }
#endif
}

